/* ----------------------------------------------------------------- */
/* --- Koordinierung von Weihnachtsmann, Elfen und Rentieren mit --- */
/* --- mit Hilfe von Semaphoren -------------------------------------*/
/* ----------------------------------------------------------------- */

/* --- Der Weihnachtsmann schlft in seinem Haus am Nordpol.     --- */
/* --- Wenn die Elfen Schwierigkeiten bei der Produktion des     --- */
/* --- Spielzeug haben, fragen sie den Weihnachtsmann um Rat.    --- */
/* --- Es mssen aber immer drei Elfen zusammenkommen, damit der --- */
/* --- Weihnachtsmann nicht dauernd gestrt wird. Wenn noch mehr --- */
/* --- Elfen Probleme haben, mssen sie warten, bis die anderen  --- */
/* --- wieder da sind.                                           --- */
/* --- Die Rentiere verbringen ihre Ferien im sonnigen Sden.    --- */
/* --- Kurz vor Weihnachten kommen sie zurck an den Nordpol.    --- */
/* --- Wenn alle neuen Rentiere da sind, ist Weihnachten, un der --- */
/* --- Weihnachtsmann spannt die Rentiere vor den Schlitten, um  --- */
/* --- seiner Arbeit nachzugehen (die Rentiere kommen immer erst --- */
/* --- im letzten Moment zurck). Die Rentiere haben immer       --- */
/* --- Vorrang vor wartenden Elfen, da Weihnachten sonst         --- */
/* --- verschoben wrde. Nach Weihnachten kehren sie zurck in   --- */
/* --- ins Warme.                                                --- */

#include "xmas.h"

//ID Semaphore
//ID shared memory
int semid;
int shmid;

//shared memory pointer
shared_t *shmPtr;

//Name des Elf
char *name;

struct sembuf sops[SEM_NUM];
union semun args;

// Semaphor freigeben
void up(int semnum)
{
    sops[0].sem_num = semnum;
    sops[0].sem_op = 1;
    sops[0].sem_flg = 0;

    if( 0 > semop(semid, &(sops[0]), 1))
    {
	perror("elf: problem with semop()");
	exit(1);
    }
}

//Semaphor anfordern
void down(int semnum)
{
    sops[0].sem_num = semnum;
    sops[0].sem_op = -1;
    sops[0].sem_flg = 0;

    if( 0 > semop(semid, &(sops[0]), 1))
    {
	perror("elf: problem with semop()");
	exit(1);
    }
}

//Signalhandler
void sighandler(int sig)
{
    printf("Signal %d caught for reindeer.\n", sig);
    
    //Semaphore freigeben
    if(0 > semctl(semid, SEM_NUM, IPC_RMID))
    {
	perror("elf: problem with semctl()");
	exit(1);
    }

    //Shared Memory ausblenden
    if(0 > shmdt(shmPtr))
    {
	perror("elf: problem with shmdt()");
	exit(1);
    }

    //Shared Memory freigeben
    if(0 > shmctl(shmid, IPC_RMID, NULL))
    {
	perror("elf: problem with shmctl()");
	exit(1);
    }

    free(name);

    exit(0);
}

int main(int argc, char* argv[])
{
    //Schlafen
    int duration;

    //Schlssel fr Semaphor
    key_t key;

    //ein Argument: Name des Elf
    if(argc != 2)
    {
	perror("elf: wrong arguments");
	exit(1);
    }

    //Name speichern
    //Achtung: strlen liefert die Anzahl der Zeichen
    //das abschlieende \0 muss auch bercksichtigt werden
    name = malloc((strlen(argv[1]) + 1) * sizeof(char));
    memcpy(name, argv[1], strlen(argv[1]) + 1); 

    //key generieren
    if(0 > (key = ftok("./README", SHM_KEY)))
    {
	perror("santa: problem with ftok()");
	exit(1);
    }
    else
	printf("Elf %s's semaphor key is %x\n", name, key);

    //Semaphor holen
    if ( 0 > (semid = semget(key,
                             SEM_NUM,
                             (IPC_CREAT | SEM_PERM))) ) {
	perror("elf: problem with semget()");
        exit(1);
    }

    //shared memory holen
    if( 0 > (shmid = shmget(SHM_KEY, sizeof(shared_t), (IPC_CREAT | SHM_PERM))))
    {
	perror("elf: problem with shmget()");
	exit(1);
    }

    //shared memory binden
    if(0 > (shmPtr = shmat(shmid, 0, 0)))
    {
	perror("elf:problem with shmat()");
	exit(1);
    }

    //Zufallszahlen initialisieren
    srandom(time(NULL) + (int)name[0] + (int)name[strlen(name)-1] + getpid());

    //auf Signale hren
    signal(SIGTERM, sighandler);
    signal(SIGINT, sighandler);

    printf("Elf %s has finished init phase\n", name);

    //Semphor da, jetzt kann der eigentliche Prozess loslegen
    while(1)
    {
	printf("Elf %s is producing toys.\n", name);

	//zur besseren Sichtbarkeit der Ablufe
	duration = 10+random()%25;
	sleep(duration);

	printf("Now elf %s has a problem.\n", name);

	//darf ich gehen?
	down(SEM_CAPACITY);
        down(SEM_MUTEX);

	printf("Elf %s may look for help.\n", name);

	//ich darf warten, bis drei da sind
	//erhhe die Anzahl der wartenden Elfen
	shmPtr->numberOfElfes++;
	//der dritte ruft den Weihnachtsmann
        if(shmPtr->numberOfElfes == ALL_ELFES)
	{
	    printf("Elf %s wakes Santa.\n", name);
	    up(SEM_MUTEX);
	    down(SEM_LEFT);
	    up(SEM_WAKEUP);
	}
	else
	{
	    //globale Variable wieder freigeben
	    up(SEM_MUTEX);
	}

	//zum Weihnachtsmann gehen
	sleep(2);
     
	//warten, bis der Weihnachtsmann ruft
	down(SEM_GO_SANTA);
	printf("Santa's calling elf %s.\n", name);

	//warten, bis der Weihnachtsmann fertig ist
	down(SEM_ADVISE_GIVEN);

	//globale Variable belegen
	down(SEM_MUTEX);
	//Elf geht nun
	shmPtr->numberOfElfes--;
	if(shmPtr->numberOfElfes == 0)
	{
	    printf("Elf %s is the last to leave.\n", name);
	    up(SEM_LEFT);
        }
	//globale Variable freigeben und bedanken
	up(SEM_MUTEX);

	duration = random()%6;
	sleep(duration);

	printf("Elf %s thanks santa for his help..\n", name);
	up(SEM_THANKS);
	
	//warten bis alle weg sind
	down(SEM_LEFT);
	up(SEM_LEFT);

	//zurckgehen
	duration = random()%3;
	sleep(duration);

	//Weg zum Weihnachtsmann freigeben
	up(SEM_CAPACITY);
       	printf("Elf %s is going back to work.\n", name);
    }
}
